/* GStreamer
 * Copyright (C) <2007> Chih-Wei Huang <cwhuang@linux.org.tw>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <string.h>

#include <gst/rtp/gstrtpbuffer.h>
#include "gstasteriskdec.h"

/* Cygwin has both netinet/in.h and winsock2.h, but
 * only one can be included, so prefer the unix one */
#ifdef HAVE_NETINET_IN_H
#  include <netinet/in.h>
#else
#ifdef HAVE_WINSOCK2_H
#  include <winsock2.h>
#endif
#endif

#define GST_ASTERISKDEC_HEADER_LEN 6

typedef struct _GstAsteriskDecHeader
{
  guint32 timestamp;            /* Timestamp */
  guint16 length;               /* Length */
} GstAsteriskDecHeader;

#define GST_ASTERISKDEC_HEADER_TIMESTAMP(buf) (((GstAsteriskDecHeader *)(GST_BUFFER_DATA (buf)))->timestamp)
#define GST_ASTERISKDEC_HEADER_LENGTH(buf) (((GstAsteriskDecHeader *)(GST_BUFFER_DATA (buf)))->length)

/* elementfactory information */
static const GstElementDetails gst_asteriskdec_details =
GST_ELEMENT_DETAILS ("RTP packet payloader",
    "Codec/Payloader/Network",
    "Payload-encodes Asterisk video file in RTP packets",
    "Chih-Wei Huang <cwhuang@linux.org.tw>");

/* Asteriskdec signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_CODEC
};

typedef enum {
  GST_ASTERISK_CODEC_H263,
  GST_ASTERISK_CODEC_H263P,
  GST_ASTERISK_CODEC_H264
} GstAsteriskCodecType;

static GstStaticPadTemplate gst_asteriskdec_src_template =
GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_H263_RTP_PAD_TEMPLATE_CAPS ";"
        GST_H263P_RTP_PAD_TEMPLATE_CAPS ";"
        GST_H264_RTP_PAD_TEMPLATE_CAPS)
    );

static GstStaticPadTemplate gst_asteriskdec_sink_template =
GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY
    );

static void gst_asteriskdec_finalize (GObject * object);

static gboolean gst_asteriskdec_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps);
static GstFlowReturn gst_asteriskdec_handle_buffer (GstBaseRTPPayload * basepayload, GstBuffer * buffer);

static void gst_asteriskdec_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_asteriskdec_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstStateChangeReturn gst_asteriskdec_change_state (GstElement *
    element, GstStateChange transition);

//static GstElementClass *parent_class = NULL;

#define GST_ASTERISK_CODEC (gst_asterisk_codec_get_type ())
static GType
gst_asterisk_codec_get_type (void)
{
  static GType gst_asterisk_codec_type = 0;
  static const GEnumValue codec_types[] = {
    {GST_ASTERISK_CODEC_H263, "Asterisk H.263 format", "h263"},
    {GST_ASTERISK_CODEC_H263P, "Asterisk H.263+ format", "h263p"},
    {GST_ASTERISK_CODEC_H264, "Asterisk H.264 format", "h264"},
    {0, NULL, NULL}
  };

  if (!gst_asterisk_codec_type)
    gst_asterisk_codec_type = g_enum_register_static ("GstAsteriskCodecType", codec_types);

  return gst_asterisk_codec_type;
}
#if 0
static GType
gst_asteriskdec_get_type (void)
{
  static GType asteriskdec_type = 0;

  if (!asteriskdec_type) {
    static const GTypeInfo asteriskdec_info = {
      sizeof (GstAsteriskDecClass),
      (GBaseInitFunc) gst_asteriskdec_base_init,
      NULL,
      (GClassInitFunc) gst_asteriskdec_class_init,
      NULL,
      NULL,
      sizeof (GstAsteriskDec),
      0,
      (GInstanceInitFunc) gst_asteriskdec_init,
    };

    asteriskdec_type =
        g_type_register_static (GST_TYPE_BASE_RTP_PAYLOAD, "GstAsteriskDec",
        &asteriskdec_info, 0);
  }
  return asteriskdec_type;
}
#endif

GST_BOILERPLATE (GstAsteriskDec, gst_asteriskdec, GstBaseRTPPayload, GST_TYPE_BASE_RTP_PAYLOAD);

static void
gst_asteriskdec_base_init (gpointer klass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_asteriskdec_src_template));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_asteriskdec_sink_template));

  gst_element_class_set_details (element_class, &gst_asteriskdec_details);
}

static void
gst_asteriskdec_class_init (GstAsteriskDecClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseRTPPayloadClass *gstbasertppayload_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;
  gstbasertppayload_class = (GstBaseRTPPayloadClass *) klass;

//  parent_class = g_type_class_peek_parent (klass);

  gobject_class->finalize = gst_asteriskdec_finalize;
  gobject_class->set_property = gst_asteriskdec_set_property;
  gobject_class->get_property = gst_asteriskdec_get_property;

  g_object_class_install_property (gobject_class, PROP_CODEC,
      g_param_spec_enum ("codec", "Codec",
          "Codec of the Asterisk file", GST_ASTERISK_CODEC,
          1, G_PARAM_READWRITE));
 
  gstelement_class->change_state = gst_asteriskdec_change_state;

  gstbasertppayload_class->set_caps = gst_asteriskdec_setcaps;
  gstbasertppayload_class->handle_buffer = gst_asteriskdec_handle_buffer;
}

static void
gst_asteriskdec_init (GstAsteriskDec * asteriskdec, GstAsteriskDecClass * klass)
{
#if 0
  asteriskdec->srcpad =
      gst_pad_new_from_static_template (&gst_asteriskdec_src_template, "src");
  gst_element_add_pad (GST_ELEMENT (asteriskdec), asteriskdec->srcpad);

  asteriskdec->sinkpad =
      gst_pad_new_from_static_template (&gst_asteriskdec_sink_template,
      "sink");
  gst_pad_set_chain_function (asteriskdec->sinkpad, gst_asteriskdec_chain);
  gst_pad_set_setcaps_function (asteriskdec->srcpad, gst_asteriskdec_setcaps);
  gst_element_add_pad (GST_ELEMENT (asteriskdec), asteriskdec->sinkpad);

  asteriskdec->adapter = gst_adapter_new ();
#endif
//gst_basertppayload_set_options (&asteriskdec->payload, "video", TRUE, "H264", 90000);
  asteriskdec->lastts = GST_CLOCK_TIME_NONE;
  asteriskdec->remainder = NULL;
}

static void
gst_asteriskdec_finalize (GObject * object)
{
  GstAsteriskDec *asteriskdec;
  asteriskdec = GST_ASTERISK_DEC (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static gboolean
gst_asteriskdec_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps)
{
  char *codecname[] = {
    "H263",
    "H263-1998",
    "H264"
  };
  GstAsteriskDec *asteriskdec = GST_ASTERISK_DEC (basepayload);

  GST_INFO_OBJECT (basepayload, "Set codec to %s\n", codecname[asteriskdec->codec]);
  gst_basertppayload_set_options (basepayload, "video", TRUE, codecname[asteriskdec->codec], 90000);
  gst_basertppayload_set_outcaps (basepayload, NULL);
//basepayload->ts_base = g_rand_int (basepayload->ts_rand);
  return TRUE;
}

static GstFlowReturn
gst_asteriskdec_push (GstBaseRTPPayload * payload, GstBuffer * buffer)
{
  GstClockTime timestamp = GST_BUFFER_TIMESTAMP (buffer);
  guint32 ts = payload->ts_base + timestamp;

  payload->timestamp = ts;
  gst_rtp_buffer_set_ssrc (buffer, payload->current_ssrc);
  gst_rtp_buffer_set_payload_type (buffer, payload->pt);
  gst_rtp_buffer_set_seq (buffer, ++payload->seqnum);
  gst_rtp_buffer_set_timestamp (buffer, ts);
  gst_buffer_set_caps (buffer, GST_PAD_CAPS (payload->srcpad));
  GST_DEBUG_OBJECT (payload, "setting RTP seqnum %d, timestamp %d", payload->seqnum, ts);

  return gst_pad_push (payload->srcpad, buffer);
}

static GstFlowReturn
gst_asteriskdec_handle_buffer (GstBaseRTPPayload * basepayload, GstBuffer * buffer)
{
  GstFlowReturn ret = GST_FLOW_OK;
  GstBuffer *buf;
  guint size;
  guint8 *data;
  GstAsteriskDec *asteriskdec = GST_ASTERISK_DEC (basepayload);

  /* initialize timestamp */
  if (!GST_CLOCK_TIME_IS_VALID(asteriskdec->lastts))
    asteriskdec->lastts = GST_BUFFER_TIMESTAMP (buffer);

  /* merge new data to previous buffer (if any) */
  buf = asteriskdec->remainder ? gst_buffer_join(asteriskdec->remainder, buffer) : buffer;
  asteriskdec->remainder = NULL;
  size = GST_BUFFER_SIZE (buf);
  data = GST_BUFFER_DATA (buf);

  while ((size > GST_ASTERISKDEC_HEADER_LEN) && (ret == GST_FLOW_OK)) {
    GstBuffer *outbuf;
    guint8 *payload;
    GstAsteriskDecHeader *asthdr = (GstAsteriskDecHeader *) data;
    guint32 samples = ntohl(asthdr->timestamp);
    guint16 payload_len = ntohs(asthdr->length), ast_len;
    gboolean marker = (payload_len & 0x8000) ? TRUE : FALSE;

    if (marker)
      payload_len &= 0x7fff;
    ast_len = GST_ASTERISKDEC_HEADER_LEN + payload_len;

    if (size < ast_len)
      break; /* not enough data */

    /* increase timestamp */
    asteriskdec->lastts += samples;
    GST_DEBUG_OBJECT (basepayload, "samples=%d payload_len=%d lastts=%d\n", samples, payload_len, asteriskdec->lastts);

    /* fill one packet */
    outbuf = gst_rtp_buffer_new_allocate (payload_len, 0, 0);
    GST_BUFFER_TIMESTAMP (outbuf) = asteriskdec->lastts;
    gst_rtp_buffer_set_marker (outbuf, marker);

    payload = gst_rtp_buffer_get_payload (outbuf);
    memcpy (payload, data + GST_ASTERISKDEC_HEADER_LEN, payload_len);
    data += ast_len, size -= ast_len;
    ret = gst_asteriskdec_push (basepayload, outbuf);
  }
  /* any remain data? */
  GST_DEBUG_OBJECT (basepayload, "left size=%d\n", size);
  if (size) {
    asteriskdec->remainder = gst_buffer_new_and_alloc(size);
    memcpy(GST_BUFFER_DATA(asteriskdec->remainder), data, size);
  }
  gst_buffer_unref (buf);

  return ret;
}

static void
gst_asteriskdec_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstAsteriskDec *asteriskdec = GST_ASTERISK_DEC (object);
  switch (prop_id) {
    case PROP_CODEC:
      asteriskdec->codec = g_value_get_enum (value);
//      gst_asteriskdec_setcaps(&asteriskdec->payload, NULL);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_asteriskdec_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstAsteriskDec *asteriskdec = GST_ASTERISK_DEC (object);
  switch (prop_id) {
    case PROP_CODEC:
      g_value_set_enum (value, asteriskdec->codec);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static GstStateChangeReturn
gst_asteriskdec_change_state (GstElement * element, GstStateChange transition)
{
  GstStateChangeReturn ret;

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      break;
    case GST_STATE_CHANGE_READY_TO_PAUSED:
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      break;
    default:
      break;
  }
  return ret;
}

gboolean
gst_asteriskdec_plugin_init (GstPlugin * plugin)
{
  return gst_element_register (plugin, "asteriskdec",
      GST_RANK_MARGINAL, GST_TYPE_ASTERISK_DEC);
}
