/* GStreamer
 * Copyright (C) <2007> Chih-Wei Huang <cwhuang@linux.org.tw>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <string.h>

#include <gst/rtp/gstrtpbuffer.h>
#include "gstasteriskenc.h"

/* Cygwin has both netinet/in.h and winsock2.h, but
 * only one can be included, so prefer the unix one */
#ifdef HAVE_NETINET_IN_H
#  include <netinet/in.h>
#else
#ifdef HAVE_WINSOCK2_H
#  include <winsock2.h>
#endif
#endif

#define GST_ASTERISKENC_HEADER_LEN 6

typedef struct _GstAsteriskEncHeader
{
  guint32 timestamp;            /* Timestamp */
  guint16 length;               /* Length */
} GstAsteriskEncHeader;

#define GST_ASTERISKENC_HEADER_TIMESTAMP(buf) (((GstAsteriskEncHeader *)(GST_BUFFER_DATA (buf)))->timestamp)
#define GST_ASTERISKENC_HEADER_LENGTH(buf) (((GstAsteriskEncHeader *)(GST_BUFFER_DATA (buf)))->length)

/* elementfactory information */
static const GstElementDetails gst_asteriskenc_details =
GST_ELEMENT_DETAILS ("RTP packet parser",
    "Codec/Depayloader/Network",
    "Extracts video from RTP and encodes in Asterisk format",
    "Chih-Wei Huang <cwhuang@linux.org.tw>");

/* Asteriskenc signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_FREQUENCY
};

static GstStaticPadTemplate gst_asteriskenc_src_template =
GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("application/x-asteriskenc"));

static GstStaticPadTemplate gst_asteriskenc_sink_template =
GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_H263_RTP_PAD_TEMPLATE_CAPS ";"
        GST_H263P_RTP_PAD_TEMPLATE_CAPS ";"
        GST_H264_RTP_PAD_TEMPLATE_CAPS)
    );

static void gst_asteriskenc_class_init (GstAsteriskEncClass * klass);
static void gst_asteriskenc_base_init (GstAsteriskEncClass * klass);
static void gst_asteriskenc_init (GstAsteriskEnc * asteriskenc);
static void gst_asteriskenc_finalize (GObject * object);

static GstFlowReturn gst_asteriskenc_chain (GstPad * pad, GstBuffer * buffer);

static void gst_asteriskenc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_asteriskenc_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstStateChangeReturn gst_asteriskenc_change_state (GstElement *
    element, GstStateChange transition);

static GstElementClass *parent_class = NULL;

static GType
gst_asteriskenc_get_type (void)
{
  static GType asteriskenc_type = 0;

  if (!asteriskenc_type) {
    static const GTypeInfo asteriskenc_info = {
      sizeof (GstAsteriskEncClass),
      (GBaseInitFunc) gst_asteriskenc_base_init,
      NULL,
      (GClassInitFunc) gst_asteriskenc_class_init,
      NULL,
      NULL,
      sizeof (GstAsteriskEnc),
      0,
      (GInstanceInitFunc) gst_asteriskenc_init,
    };

    asteriskenc_type =
        g_type_register_static (GST_TYPE_ELEMENT, "GstAsteriskEnc",
        &asteriskenc_info, 0);
  }
  return asteriskenc_type;
}

static void
gst_asteriskenc_base_init (GstAsteriskEncClass * klass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_asteriskenc_src_template));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_asteriskenc_sink_template));

  gst_element_class_set_details (element_class, &gst_asteriskenc_details);
}

static void
gst_asteriskenc_class_init (GstAsteriskEncClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  parent_class = g_type_class_peek_parent (klass);

  gobject_class->finalize = gst_asteriskenc_finalize;

  gobject_class->set_property = gst_asteriskenc_set_property;
  gobject_class->get_property = gst_asteriskenc_get_property;

  gstelement_class->change_state = gst_asteriskenc_change_state;
}
#if 0
static gboolean
gst_asteriskenc_setcaps (GstPad * pad, GstCaps * caps)
{
  GstAsteriskEnc *asteriskenc = GST_ASTERISK_ENC (GST_OBJECT_PARENT (pad));
  GstStructure *structure = gst_caps_get_structure (caps, 0);
  const gchar *codec = gst_structure_get_string (structure, "encoding-name");

  if (!codec)
    goto wrong_caps;

  gboolean ret = gst_pad_set_caps (asteriskenc->sinkpad, caps);
  return ret;

  wrong_caps:
  return ret;
}
#endif
static void
gst_asteriskenc_init (GstAsteriskEnc * asteriskenc)
{
  asteriskenc->srcpad =
      gst_pad_new_from_static_template (&gst_asteriskenc_src_template, "src");
  gst_element_add_pad (GST_ELEMENT (asteriskenc), asteriskenc->srcpad);

  asteriskenc->sinkpad =
      gst_pad_new_from_static_template (&gst_asteriskenc_sink_template,
      "sink");
  gst_pad_set_chain_function (asteriskenc->sinkpad, gst_asteriskenc_chain);
#if 0
  gst_pad_set_setcaps_function (asteriskenc->srcpad, gst_asteriskenc_setcaps);
#endif
  gst_element_add_pad (GST_ELEMENT (asteriskenc), asteriskenc->sinkpad);

  asteriskenc->adapter = gst_adapter_new ();
  asteriskenc->lastts = GST_CLOCK_TIME_NONE;
}

static void
gst_asteriskenc_finalize (GObject * object)
{
  GstAsteriskEnc *asteriskenc = GST_ASTERISK_ENC (object);

  g_object_unref (asteriskenc->adapter);
  asteriskenc->adapter = NULL;

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static GstFlowReturn
gst_asteriskenc_chain (GstPad * pad, GstBuffer * buf)
{
  GstBuffer *outbuf;
  GstFlowReturn ret;
  GstAsteriskEnc *asteriskenc = GST_ASTERISK_ENC (GST_OBJECT_PARENT (pad));

  if (!gst_rtp_buffer_validate (buf))
    goto bad_packet;

  {
    gint payload_len;
    guint8 *payload;
    gboolean M;
    guint32 timestamp;
    guint32 samples;
    guint16 asterisk_len;

    payload_len = gst_rtp_buffer_get_payload_len (buf);
    payload = gst_rtp_buffer_get_payload (buf);

    M = gst_rtp_buffer_get_marker (buf);
    timestamp = gst_rtp_buffer_get_timestamp (buf);
    outbuf = gst_buffer_new_and_alloc (payload_len +
        GST_ASTERISKENC_HEADER_LEN);

    /* build the asterisk header */
    asterisk_len = payload_len;
    if (M)
      asterisk_len |= 0x8000;
    samples = GST_CLOCK_TIME_IS_VALID(asteriskenc->lastts) ? timestamp - asteriskenc->lastts : 0;
    asteriskenc->lastts = timestamp;

    GST_ASTERISKENC_HEADER_TIMESTAMP (outbuf) = htonl (samples);
    GST_ASTERISKENC_HEADER_LENGTH (outbuf) = htons (asterisk_len);

    /* copy the data into place */
    memcpy (GST_BUFFER_DATA (outbuf) + GST_ASTERISKENC_HEADER_LEN, payload,
        payload_len);

    GST_BUFFER_TIMESTAMP (outbuf) = timestamp;
    gst_buffer_set_caps (outbuf,
        (GstCaps *) gst_pad_get_pad_template_caps (asteriskenc->srcpad));

    ret = gst_pad_push (asteriskenc->srcpad, outbuf);

    gst_buffer_unref (buf);
  }

  return ret;

bad_packet:
  {
    GST_DEBUG ("Packet does not validate");
    gst_buffer_unref (buf);
    return GST_FLOW_ERROR;
  }
}

static void
gst_asteriskenc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  //GstAsteriskEnc *asteriskenc = GST_ASTERISK_ENC (object);

  switch (prop_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_asteriskenc_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  //GstAsteriskEnc *asteriskenc = GST_ASTERISK_ENC (object);

  switch (prop_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static GstStateChangeReturn
gst_asteriskenc_change_state (GstElement * element, GstStateChange transition)
{
  GstStateChangeReturn ret;
  GstAsteriskEnc *asteriskenc = GST_ASTERISK_ENC (element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      break;
    case GST_STATE_CHANGE_READY_TO_PAUSED:
      gst_adapter_clear (asteriskenc->adapter);
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      break;
    default:
      break;
  }
  return ret;
}

gboolean
gst_asteriskenc_plugin_init (GstPlugin * plugin)
{
  return gst_element_register (plugin, "asteriskenc",
      GST_RANK_MARGINAL, GST_TYPE_ASTERISK_ENC);
}
